import logging
import os
import warnings

import numpy as np
from astropy import cosmology as cosmo
from autofit import conf
from autofit.tools import phase
from autofit.tools.phase_property import PhasePropertyCollection
from autofit.optimize import non_linear

from autolens import exc
from autolens.data.array import mask as msk
from autolens.data.plotters import ccd_plotters
from autolens.lens import lens_data as li, lens_fit
from autolens.lens import ray_tracing
from autolens.lens import sensitivity_fit
from autolens.lens.plotters import sensitivity_fit_plotters, ray_tracing_plotters, lens_fit_plotters
from autolens.model.galaxy import galaxy as g, galaxy_model as gm, galaxy_fit, galaxy_data as gd
from autolens.model.galaxy.plotters import galaxy_fit_plotters

logger = logging.getLogger(__name__)
logger.level = logging.DEBUG


def default_mask_function(image):
    return msk.Mask.circular(shape=image.shape, pixel_scale=image.pixel_scale, radius_arcsec=3.0)

def setup_phase_mask(data, mask, mask_function, inner_circular_mask_radii):

    if mask_function is not None:
        mask = mask_function(image=data.image)
    elif mask is None and mask_function is None:
        mask = default_mask_function(image=data.image)

    if inner_circular_mask_radii is not None:
        inner_mask = msk.Mask.circular(shape=mask.shape, pixel_scale=mask.pixel_scale,
                                       radius_arcsec=inner_circular_mask_radii, invert=True)
        mask = mask + inner_mask

    return mask

class ResultsCollection(list):
    def __init__(self, results):
        super().__init__(results)

    @property
    def last(self):
        if len(self) > 0:
            return self[-1]
        return None

    @property
    def first(self):
        if len(self) > 0:
            return self[0]
        return None


class AbstractPhase(phase.AbstractPhase):

    def __init__(self, phase_name, optimizer_class=non_linear.MultiNest, cosmology=cosmo.Planck15,
                 auto_link_priors=False):
        """
        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        phase_name: str
            The name of this phase
        """
        self.optimizer = optimizer_class(name=phase_name)
        self.cosmology = cosmology
        self.phase_name = phase_name
        self.auto_link_priors = auto_link_priors

    @property
    def constant(self):
        """
        Convenience method

        Returns
        -------
        ModelInstance
            A model instance comprising all the constant objects in this lens
        """
        return self.optimizer.constant

    @property
    def variable(self):
        """
        Convenience method

        Returns
        -------
        ModelMapper
            A model mapper comprising all the variable (prior) objects in this lens
        """
        return self.optimizer.variable

    @property
    def galaxy_model_tuples(self):
        """
        Returns
        -------
        galaxy_model_tuples: [(String, GalaxyModel)]
            A list of tuples containing galaxy model names and instances.
        """
        return [tup for tup in self.optimizer.variable.prior_model_tuples if
                isinstance(tup.prior_model, gm.GalaxyModel)]

    def match_instance_to_models(self, instance):
        """
        Matches named galaxies associated with the instance to named galaxies associated with this phase.

        Parameters
        ----------
        instance: ModelInstance
            An instance with named galaxy attributes.

        Returns
        -------
        tuples: [(String, Galaxy, GalaxyModel)]
            A list of tuples associating galaxy instances from the model instance object with galaxy models in this
            phase.
        """
        galaxy_dict = dict(instance.name_instance_tuples_for_class(g.Galaxy))
        return [(key, galaxy_dict[key], value) for key, value in self.galaxy_model_tuples if key in galaxy_dict]

    def fit_priors(self, instance, fitting_function):
        """
        Update the priors in this phase by fitting each galaxy model to a galaxy with the same name from a previous
        phase if such a galaxy exists.

        Parameters
        ----------
        instance: ModelInstance
            An object with named galaxy attributes
        fitting_function: (Galaxy, GalaxyModel) -> GalaxyModel
            A function that takes a galaxy and a galaxy model and returns a GalaxyModel produced by combining a best fit
            between the original galaxy and galaxy model with prior widths given by the configuration.
        """
        tuples = self.match_instance_to_models(instance)
        for t in tuples:
            name = t[0]
            galaxy = t[1]
            galaxy_model = t[2]
            new_galaxy_model = fitting_function(galaxy, galaxy_model)
            for phase_property_collection in self.phase_property_collections:
                if hasattr(phase_property_collection, name):
                    setattr(phase_property_collection, name, new_galaxy_model)

    def fit_priors_with_results(self, results, fitting_function):
        """
        Update the priors in this phase by fitting each galaxy model to a galaxy with the same name from a previous
        phase if such a galaxy exists.

        Results later in the list take precedence, with the last instance of any galaxies that share a name being kept.

        Parameters
        ----------
        results: [Results]
            A list of results from previous phases.
        fitting_function: (Galaxy, GalaxyModel) -> GalaxyModel
            A function that takes a galaxy and a galaxy model and returns a GalaxyModel produced by combining a best fit
            between the original galaxy and galaxy model with prior widths given by the configuration.
        """
        if results is not None and len(results) > 0:
            instances = [r.constant for r in results]
            instance = instances[0]
            for next_instance in instances[1:]:
                instance += next_instance
            self.fit_priors(instance, fitting_function)

    @property
    def phase_property_collections(self):
        """
        Returns
        -------
        phase_property_collections: [PhasePropertyCollection]
            A list of phase property collections associated with this phase. This is used in automated prior passing and
            should be overridden for any phase that contains its own PhasePropertyCollections.
        """
        return []

    @property
    def path(self):
        return self.optimizer.path

    @property
    def doc(self):
        if self.__doc__ is not None:
            return self.__doc__.replace("  ", "").replace("\n", " ")

    def pass_priors(self, previous_results):
        """
        Perform any prior or constant passing. This could involve setting model attributes equal to priors or constants
        from a previous phase.

        Parameters
        ----------
        previous_results: ResultsCollection
            The result of the previous phase
        """
        pass

    # noinspection PyAbstractClass
    class Analysis(non_linear.Analysis):

        def __init__(self, cosmology, phase_name, previous_results=None):
            """
            An lens object

            Parameters
            ----------
            phase_name: str
                The name of the phase to which this analysis belongs
            previous_results: ResultsCollection
                The results of all previous phases
            """

            self.previous_results = previous_results
            self.cosmology = cosmology
            self.phase_name = phase_name
            self.phase_output_path = "{}/{}".format(conf.instance.output_path, self.phase_name)

            log_file = conf.instance.general.get('output', 'log_file', str)
            if not len(log_file.replace(" ", "")) == 0:
                log_path = "{}/{}".format(self.phase_output_path, log_file)
                logger.handlers = [logging.FileHandler(log_path)]
                logger.propagate = False

            self.position_threshold = conf.instance.general.get('positions', 'position_threshold', float)
            self.plot_count = 0
            self.output_image_path = "{}/image/".format(self.phase_output_path)
            make_path_if_does_not_exist(path=self.output_image_path)
            self.output_fits_path = "{}/image/fits/".format(self.phase_output_path)
            make_path_if_does_not_exist(path=self.output_fits_path)

        @property
        def last_results(self):
            if self.previous_results is not None:
                return self.previous_results.last

        def tracer_for_instance(self, instance):
            raise NotImplementedError()

        def padded_tracer_for_instance(self, instance):
            raise NotImplementedError()

        def fit_for_tracers(self, tracer, padded_tracer):
            raise NotImplementedError()

        def figure_of_merit_for_fit(self, tracer):
            raise NotImplementedError()

    def make_result(self, result, analysis):
        return self.__class__.Result(constant=result.constant, figure_of_merit=result.figure_of_merit,
                                     variable=result.variable, analysis=analysis, optimizer=self.optimizer)

    class Result(non_linear.Result):

        def __init__(self, constant, figure_of_merit, variable, analysis, optimizer):
            """
            The result of a phase
            """
            super(Phase.Result, self).__init__(constant=constant, figure_of_merit=figure_of_merit, variable=variable)

            self.analysis = analysis
            self.optimizer = optimizer

        @property
        def most_likely_tracer(self):
            return self.analysis.tracer_for_instance(instance=self.constant)

        @property
        def most_likely_padded_tracer(self):
            return self.analysis.padded_tracer_for_instance(instance=self.constant)

        @property
        def most_likely_fit(self):
            return self.analysis.fit_for_tracers(tracer=self.most_likely_tracer,
                                                 padded_tracer=self.most_likely_padded_tracer)

        @property
        def unmasked_model_image(self):
            return self.most_likely_fit.unmasked_model_image

        @property
        def unmasked_model_image_of_planes(self):
            return self.most_likely_fit.unmasked_model_image_of_planes

        @property
        def unmasked_model_image_of_planes_and_galaxies(self):
            return self.most_likely_fit.unmasked_model_image_of_planes_and_galaxies


class Phase(AbstractPhase):

    def run(self, image, previous_results=None, mask=None):
        raise NotImplementedError()

    # noinspection PyAbstractClass
    class Analysis(AbstractPhase.Analysis):

        def __init__(self, cosmology, phase_name, previous_results=None):

            super(Phase.Analysis, self).__init__(cosmology=cosmology, phase_name=phase_name,
                                                 previous_results=previous_results)

            self.should_plot_mask = \
                conf.instance.general.get('output', 'plot_mask_on_images', bool)
            self.extract_array_from_mask = \
                conf.instance.general.get('output', 'extract_images_from_mask', bool)
            self.zoom_around_mask = \
                conf.instance.general.get('output', 'zoom_around_mask_of_images', bool)
            self.should_plot_positions = \
                conf.instance.general.get('output', 'plot_positions_on_images', bool)
            self.plot_units = \
                conf.instance.general.get('output', 'plot_units', str).strip()

            self.plot_ray_tracing_all_at_end_png = \
                conf.instance.general.get('output', 'plot_ray_tracing_all_at_end_png', bool)
            self.plot_ray_tracing_all_at_end_fits = \
                conf.instance.general.get('output', 'plot_ray_tracing_all_at_end_fits', bool)

            self.plot_ray_tracing_as_subplot = \
                conf.instance.general.get('output', 'plot_ray_tracing_as_subplot', bool)
            self.plot_ray_tracing_image_plane_image = \
                conf.instance.general.get('output', 'plot_ray_tracing_image_plane_image', bool)
            self.plot_ray_tracing_source_plane = \
                conf.instance.general.get('output', 'plot_ray_tracing_source_plane_image', bool)
            self.plot_ray_tracing_surface_density = \
                conf.instance.general.get('output', 'plot_ray_tracing_surface_density', bool)
            self.plot_ray_tracing_potential = \
                conf.instance.general.get('output', 'plot_ray_tracing_potential', bool)
            self.plot_ray_tracing_deflections = \
                conf.instance.general.get('output', 'plot_ray_tracing_deflections', bool)


class PhasePositions(AbstractPhase):
    lens_galaxies = PhasePropertyCollection("lens_galaxies")

    @property
    def phase_property_collections(self):
        return [self.lens_galaxies]

    def __init__(self, phase_name, lens_galaxies=None, optimizer_class=non_linear.MultiNest, cosmology=cosmo.Planck15,
                 auto_link_priors=False):
        super().__init__(optimizer_class=optimizer_class, cosmology=cosmology,
                         phase_name=phase_name, auto_link_priors=auto_link_priors)
        self.lens_galaxies = lens_galaxies

    def run(self, positions, pixel_scale, previous_results=None):
        """
        Run this phase.

        Parameters
        ----------
        pixel_scale
        positions
        previous_results: ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper.
        """
        analysis = self.make_analysis(positions=positions, pixel_scale=pixel_scale, previous_results=previous_results)
        result = self.run_analysis(analysis)
        return self.make_result(result, analysis)

    def make_analysis(self, positions, pixel_scale, previous_results=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        pixel_scale
        positions
        previous_results: ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens: Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """
        self.pass_priors(previous_results)
        analysis = self.__class__.Analysis(positions=positions, pixel_scale=pixel_scale, cosmology=self.cosmology,
                                           phase_name=self.phase_name, previous_results=previous_results)
        return analysis

    # noinspection PyAbstractClass
    class Analysis(Phase.Analysis):

        def __init__(self, positions, pixel_scale, cosmology, phase_name, previous_results=None):
            super().__init__(cosmology=cosmology, phase_name=phase_name, previous_results=previous_results)

            self.positions = list(map(lambda position_set: np.asarray(position_set), positions))
            self.pixel_scale = pixel_scale

        def visualize(self, instance, suffix, during_analysis):
            pass

        def fit(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit: Fit
                A fractional value indicating how well this model fit and the model lens_data itself
            """
            tracer = self.tracer_for_instance(instance)
            fit = self.fit_for_tracer(tracer)
            return fit.figure_of_merit

        def tracer_for_instance(self, instance):
            return ray_tracing.TracerImageSourcePlanesPositions(lens_galaxies=instance.lens_galaxies,
                                                                image_plane_positions=self.positions,
                                                                cosmology=self.cosmology)

        def fit_for_tracer(self, tracer):
            return lens_fit.LensPositionFit(positions=tracer.source_plane.positions, noise_map=self.pixel_scale)

        @classmethod
        def log(cls, instance):
            logger.debug(
                "\nRunning lens lens for... \n\nLens Galaxy::\n{}\n\n".format(instance.lens_galaxies))


class PhaseImaging(Phase):

    def __init__(self, phase_name, optimizer_class=non_linear.MultiNest, sub_grid_size=2, image_psf_shape=None,
                 pixelization_psf_shape=None, use_positions=False, mask_function=None, inner_circular_mask_radii=None,
                 cosmology=cosmo.Planck15, auto_link_priors=False):

        """

        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        super(PhaseImaging, self).__init__(optimizer_class=optimizer_class, cosmology=cosmology,
                                           phase_name=phase_name, auto_link_priors=auto_link_priors)
        self.sub_grid_size = sub_grid_size
        self.image_psf_shape = image_psf_shape
        self.pixelization_psf_shape = pixelization_psf_shape
        self.use_positions = use_positions
        self.mask_function = mask_function
        self.inner_circular_mask_radii = inner_circular_mask_radii

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def modify_image(self, image, previous_results):
        """
        Customize an lens_data. e.g. removing lens light.

        Parameters
        ----------
        image: scaled_array.ScaledSquarePixelArray
            An lens_data that has been masked
        previous_results: ResultsCollection
            The result of the previous lens

        Returns
        -------
        lens_data: scaled_array.ScaledSquarePixelArray
            The modified image (not changed by default)
        """
        return image

    def run(self, data, previous_results=None, mask=None, positions=None):
        """
        Run this phase.

        Parameters
        ----------
        mask: Mask
            The default masks passed in by the pipeline
        previous_results: ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        data: scaled_array.ScaledSquarePixelArray
            An lens_data that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper.
        """
        analysis = self.make_analysis(data=data, previous_results=previous_results, mask=mask, positions=positions)

        result = self.run_analysis(analysis)

        return self.make_result(result, analysis)

    def make_analysis(self, data, previous_results=None, mask=None, positions=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        mask: Mask
            The default masks passed in by the pipeline
        data: im.CCD
            An lens_data that has been masked
        previous_results: ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """

        mask = setup_phase_mask(data=data, mask=mask, mask_function=self.mask_function,
                                inner_circular_mask_radii=self.inner_circular_mask_radii)

        if self.use_positions and positions is not None:
            positions = list(map(lambda position_set: np.asarray(position_set), positions))
        elif not self.use_positions:
            positions = None
        elif self.use_positions and positions is None:
            raise exc.PhaseException('You have specified for a phase to use positions, but not input positions to the '
                                     'pipeline when you ran it.')

        lens_data = li.LensData(ccd_data=data, mask=mask, sub_grid_size=self.sub_grid_size,
                                image_psf_shape=self.image_psf_shape, positions=positions)

        modified_image = self.modify_image(image=lens_data.image, previous_results=previous_results)
        lens_data = lens_data.new_lens_data_with_modified_image(modified_image=modified_image)

        self.pass_priors(previous_results)

        self.output_phase_info()

        analysis = self.__class__.Analysis(lens_data=lens_data, cosmology=self.cosmology,
                                           phase_name=self.phase_name, previous_results=previous_results)
        return analysis

    def output_phase_info(self):

        file_phase_info = "{}/{}/{}".format(conf.instance.output_path, self.phase_name, 'phase.info')

        with open(file_phase_info, 'w') as phase_info:

            phase_info.write('Optimizer = {} \n'.format(type(self.optimizer).__name__))
            phase_info.write('Sub-grid size = {} \n'.format(self.sub_grid_size))
            phase_info.write('Image PSF shape = {} \n'.format(self.image_psf_shape))
            phase_info.write('Pixelization PSF shape = {} \n'.format(self.pixelization_psf_shape))
            phase_info.write('Use positions = {} \n'.format(self.use_positions))
            position_threshold = conf.instance.general.get('positions', 'position_threshold', float)
            phase_info.write('Positions Threshold = {} \n'.format(position_threshold))
            phase_info.write('Cosmology = {} \n'.format(self.cosmology))
            phase_info.write('Auto Link Priors = {} \n'.format(self.auto_link_priors))

            phase_info.close()

    # noinspection PyAbstractClass
    class Analysis(Phase.Analysis):

        def __init__(self, lens_data, cosmology, phase_name, previous_results=None):

            super(PhaseImaging.Analysis, self).__init__(cosmology=cosmology, phase_name=phase_name,
                                                        previous_results=previous_results)

            self.lens_data = lens_data

            self.should_plot_image_plane_pix = \
                conf.instance.general.get('output', 'plot_image_plane_adaptive_pixelization_grid', bool)

            self.plot_data_as_subplot =\
                conf.instance.general.get('output', 'plot_data_as_subplot', bool)
            self.plot_data_image = \
                conf.instance.general.get('output', 'plot_data_image', bool)
            self.plot_data_noise_map = \
                conf.instance.general.get('output', 'plot_data_noise_map', bool)
            self.plot_data_psf = \
                conf.instance.general.get('output', 'plot_data_psf', bool)
            self.plot_data_signal_to_noise_map = \
                conf.instance.general.get('output', 'plot_data_signal_to_noise_map', bool)

            self.plot_lens_fit_all_at_end_png = \
                conf.instance.general.get('output', 'plot_lens_fit_all_at_end_png', bool)
            self.plot_lens_fit_all_at_end_fits = \
                conf.instance.general.get('output', 'plot_lens_fit_all_at_end_fits', bool)

            self.plot_lens_fit_as_subplot = \
                conf.instance.general.get('output', 'plot_lens_fit_as_subplot', bool)
            self.plot_lens_fit_image = \
                conf.instance.general.get('output', 'plot_lens_fit_image', bool)
            self.plot_lens_fit_noise_map = \
                conf.instance.general.get('output', 'plot_lens_fit_noise_map', bool)
            self.plot_lens_fit_signal_to_noise_map = \
                conf.instance.general.get('output', 'plot_lens_fit_signal_to_noise_map', bool)
            self.plot_lens_fit_lens_subtracted_image = \
                conf.instance.general.get('output', 'plot_lens_fit_lens_subtracted_image', bool)
            self.plot_lens_fit_model_image = \
                conf.instance.general.get('output', 'plot_lens_fit_model_image', bool)
            self.plot_lens_fit_lens_model_image = \
                conf.instance.general.get('output', 'plot_lens_fit_lens_model_image', bool)
            self.plot_lens_fit_source_model_image = \
                conf.instance.general.get('output', 'plot_lens_fit_source_model_image', bool)
            self.plot_lens_fit_source_plane_image = \
                conf.instance.general.get('output', 'plot_lens_fit_source_plane_image', bool)
            self.plot_lens_fit_residual_map = \
                conf.instance.general.get('output', 'plot_lens_fit_residual_map', bool)
            self.plot_lens_fit_chi_squared_map = \
                conf.instance.general.get('output', 'plot_lens_fit_chi_squared_map', bool)
            self.plot_lens_fit_contribution_map = \
                conf.instance.general.get('output', 'plot_lens_fit_contribution_map', bool)

        def fit(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit : Fit
                A fractional value indicating how well this model fit and the model lens_data itself
            """
            self.check_positions_trace_within_threshold(instance)
            tracer = self.tracer_for_instance(instance)
            fit = self.fit_for_tracers(tracer=tracer, padded_tracer=None)
            return fit.figure_of_merit

        def visualize(self, instance, suffix, during_analysis):

            self.plot_count += 1

            if self.should_plot_mask:
                mask = self.lens_data.mask
            else:
                mask = None

            if self.should_plot_positions:
                positions = self.lens_data.positions
            else:
                positions = None

            if self.plot_data_as_subplot:

                ccd_plotters.plot_ccd_subplot(
                    ccd_data=self.lens_data.ccd_data, mask=mask, extract_array_from_mask=self.extract_array_from_mask,
                    zoom_around_mask=self.zoom_around_mask, positions=positions,
                    units=self.plot_units,
                    output_path=self.output_image_path, output_format='png')

            ccd_plotters.plot_ccd_individual(
                ccd_data=self.lens_data.ccd_data, mask=mask, extract_array_from_mask=self.extract_array_from_mask,
                zoom_around_mask=self.zoom_around_mask, positions=positions,
                should_plot_image=self.plot_data_image,
                should_plot_noise_map=self.plot_data_noise_map,
                should_plot_psf=self.plot_data_psf,
                should_plot_signal_to_noise_map=self.plot_data_signal_to_noise_map,
                units=self.plot_units,
                output_path=self.output_image_path, output_format='png')


            tracer = self.tracer_for_instance(instance)
            padded_tracer = self.padded_tracer_for_instance(instance)

            if self.plot_ray_tracing_as_subplot:

                ray_tracing_plotters.plot_ray_tracing_subplot(
                    tracer=tracer, mask=mask, extract_array_from_mask=self.extract_array_from_mask,
                    zoom_around_mask=self.zoom_around_mask, positions=positions,
                    units=self.plot_units,
                    output_path=self.output_image_path, output_format='png')

            fit = self.fit_for_tracers(tracer=tracer, padded_tracer=padded_tracer)

            if self.plot_lens_fit_as_subplot:

                lens_fit_plotters.plot_fit_subplot(
                    fit=fit, should_plot_mask=self.should_plot_mask,
                    extract_array_from_mask=self.extract_array_from_mask, zoom_around_mask=self.zoom_around_mask,
                    positions=positions, should_plot_image_plane_pix=self.should_plot_image_plane_pix,
                    units=self.plot_units,
                    output_path=self.output_image_path, output_format='png')

            if during_analysis:

                ray_tracing_plotters.plot_ray_tracing_individual(
                    tracer=tracer, mask=mask, extract_array_from_mask=self.extract_array_from_mask,
                    zoom_around_mask=self.zoom_around_mask, positions=positions,
                    should_plot_image_plane_image=self.plot_ray_tracing_image_plane_image,
                    should_plot_source_plane=self.plot_ray_tracing_source_plane,
                    should_plot_surface_density=self.plot_ray_tracing_surface_density,
                    should_plot_potential=self.plot_ray_tracing_potential,
                    should_plot_deflections=self.plot_ray_tracing_deflections,
                    units=self.plot_units,
                    output_path=self.output_image_path, output_format='png')

                lens_fit_plotters.plot_fit_individuals(
                    fit=fit, should_plot_mask=self.should_plot_mask,
                    extract_array_from_mask=self.extract_array_from_mask,zoom_around_mask=self.zoom_around_mask,
                    positions=positions, should_plot_image_plane_pix=self.should_plot_image_plane_pix,
                    should_plot_image=self.plot_lens_fit_image,
                    should_plot_noise_map=self.plot_lens_fit_noise_map,
                    should_plot_signal_to_noise_map=self.plot_lens_fit_signal_to_noise_map,
                    should_plot_lens_subtracted_image=self.plot_lens_fit_lens_subtracted_image,
                    should_plot_model_image=self.plot_lens_fit_model_image,
                    should_plot_lens_model_image=self.plot_lens_fit_lens_model_image,
                    should_plot_source_model_image=self.plot_lens_fit_source_model_image,
                    should_plot_source_plane_image=self.plot_lens_fit_source_plane_image,
                    should_plot_residual_map=self.plot_lens_fit_residual_map,
                    should_plot_chi_squared_map=self.plot_lens_fit_chi_squared_map,
                    units=self.plot_units,
                    output_path=self.output_image_path, output_format='png')

            elif not during_analysis:

                if self.plot_ray_tracing_all_at_end_png:

                    ray_tracing_plotters.plot_ray_tracing_individual(
                        tracer=tracer, mask=mask, extract_array_from_mask=self.extract_array_from_mask,
                        zoom_around_mask=self.zoom_around_mask, positions=positions,
                        should_plot_image_plane_image=True,
                        should_plot_source_plane=True,
                        should_plot_surface_density=True,
                        should_plot_potential=True,
                        should_plot_deflections=True,
                        units=self.plot_units,
                        output_path=self.output_image_path, output_format='png')

                if self.plot_ray_tracing_all_at_end_fits:

                    ray_tracing_plotters.plot_ray_tracing_individual(
                        tracer=tracer, mask=mask, extract_array_from_mask=self.extract_array_from_mask,
                        zoom_around_mask=self.zoom_around_mask, positions=positions,
                        should_plot_image_plane_image=True,
                        should_plot_source_plane=True,
                        should_plot_surface_density=True,
                        should_plot_potential=True,
                        should_plot_deflections=True,
                        output_path=self.output_fits_path, output_format='fits')

                if self.plot_lens_fit_all_at_end_png:

                    lens_fit_plotters.plot_fit_individuals(
                        fit=fit, should_plot_mask=self.should_plot_mask,
                        extract_array_from_mask=self.extract_array_from_mask, zoom_around_mask=self.zoom_around_mask,
                        positions=positions, should_plot_image_plane_pix=self.should_plot_image_plane_pix,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_signal_to_noise_map=True,
                        should_plot_lens_subtracted_image=True,
                        should_plot_model_image=True,
                        should_plot_lens_model_image=True,
                        should_plot_source_model_image=True,
                        should_plot_source_plane_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=self.output_image_path, output_format='png')

                if self.plot_lens_fit_all_at_end_fits:

                    lens_fit_plotters.plot_fit_individuals(
                        fit=fit, should_plot_mask=self.should_plot_mask,
                        extract_array_from_mask=self.extract_array_from_mask, zoom_around_mask=self.zoom_around_mask,
                        positions=positions, should_plot_image_plane_pix=self.should_plot_image_plane_pix,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_signal_to_noise_map=True,
                        should_plot_lens_subtracted_image=True,
                        should_plot_model_image=True,
                        should_plot_lens_model_image=True,
                        should_plot_source_model_image=True,
                        should_plot_source_plane_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        output_path=self.output_fits_path, output_format='fits')

            return fit

        def fit_for_tracers(self, tracer, padded_tracer):
            return lens_fit.fit_lens_data_with_tracer(lens_data=self.lens_data, tracer=tracer,
                                                      padded_tracer=padded_tracer)

        def check_positions_trace_within_threshold(self, instance):

            if self.lens_data.positions is not None:

                tracer = ray_tracing.TracerImageSourcePlanesPositions(lens_galaxies=instance.lens_galaxies,
                                                                      image_plane_positions=self.lens_data.positions)
                fit = lens_fit.LensPositionFit(positions=tracer.source_plane.positions,
                                               noise_map=self.lens_data.pixel_scale)

                if not fit.maximum_separation_within_threshold(self.position_threshold):
                    return exc.RayTracingException

        def map_to_1d(self, data):
            """Convenience method"""
            return self.lens_data.mask.map_2d_array_to_masked_1d_array(data)

    class Result(Phase.Result):

        def __init__(self, constant, figure_of_merit, variable, analysis, optimizer):
            """
            The result of a phase
            """
            super(PhaseImaging.Result, self).__init__(constant=constant, figure_of_merit=figure_of_merit,
                                                      variable=variable, analysis=analysis, optimizer=optimizer)


class LensPlanePhase(PhaseImaging):
    """
    Fit only the lens galaxy light.
    """

    lens_galaxies = PhasePropertyCollection("lens_galaxies")

    @property
    def phase_property_collections(self):
        return [self.lens_galaxies]

    def __init__(self, phase_name, lens_galaxies=None, optimizer_class=non_linear.MultiNest, sub_grid_size=2,
                 image_psf_shape=None, mask_function=None, inner_circular_mask_radii=None, cosmology=cosmo.Planck15,
                 auto_link_priors=False):
        super(LensPlanePhase, self).__init__(optimizer_class=optimizer_class,
                                             sub_grid_size=sub_grid_size,
                                             image_psf_shape=image_psf_shape,
                                             mask_function=mask_function,
                                             inner_circular_mask_radii=inner_circular_mask_radii,
                                             cosmology=cosmology,
                                             phase_name=phase_name,
                                             auto_link_priors=auto_link_priors)
        self.lens_galaxies = lens_galaxies

    class Analysis(PhaseImaging.Analysis):

        def __init__(self, lens_data, cosmology, phase_name, previous_results=None):
            super(LensPlanePhase.Analysis, self).__init__(lens_data=lens_data, cosmology=cosmology,
                                                          phase_name=phase_name, previous_results=previous_results)

        def tracer_for_instance(self, instance):
            return ray_tracing.TracerImagePlane(lens_galaxies=instance.lens_galaxies,
                                                image_plane_grid_stack=self.lens_data.grid_stack,
                                                cosmology=self.cosmology)

        def padded_tracer_for_instance(self, instance):
            return ray_tracing.TracerImagePlane(lens_galaxies=instance.lens_galaxies,
                                                image_plane_grid_stack=self.lens_data.padded_grid_stack,
                                                cosmology=self.cosmology)

        @classmethod
        def log(cls, instance):
            logger.debug(
                "\nRunning lens lens for... \n\nLens Galaxy::\n{}\n\n".format(instance.lens_galaxies))

    class Result(PhaseImaging.Result):

        def __init__(self, constant, figure_of_merit, variable, analysis, optimizer):
            """
            The result of a phase
            """

            super(LensPlanePhase.Result, self).__init__(constant=constant, figure_of_merit=figure_of_merit,
                                                        variable=variable, analysis=analysis, optimizer=optimizer)

        @property
        def unmasked_lens_plane_model_image(self):
            return self.most_likely_fit.unmasked_model_image_of_planes[0]


class LensSourcePlanePhase(PhaseImaging):
    """
    Fit a simple source and lens system.
    """

    lens_galaxies = PhasePropertyCollection("lens_galaxies")
    source_galaxies = PhasePropertyCollection("source_galaxies")

    @property
    def phase_property_collections(self):
        return [self.lens_galaxies, self.source_galaxies]

    def __init__(self, phase_name, lens_galaxies=None, source_galaxies=None, optimizer_class=non_linear.MultiNest,
                 sub_grid_size=2, image_psf_shape=None, use_positions=False, mask_function=None,
                 inner_circular_mask_radii=None, cosmology=cosmo.Planck15, auto_link_priors=False):
        """
        A phase with a simple source/lens model

        Parameters
        ----------
        lens_galaxies : [g.Galaxy] | [gm.GalaxyModel]
            A galaxy that acts as a gravitational lens
        source_galaxies: [g.Galaxy] | [gm.GalaxyModel]
            A galaxy that is being lensed
        optimizer_class: class
            The class of a non-linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        super(LensSourcePlanePhase, self).__init__(optimizer_class=optimizer_class,
                                                   sub_grid_size=sub_grid_size,
                                                   image_psf_shape=image_psf_shape,
                                                   use_positions=use_positions,
                                                   mask_function=mask_function,
                                                   inner_circular_mask_radii=inner_circular_mask_radii,
                                                   cosmology=cosmology,
                                                   phase_name=phase_name,
                                                   auto_link_priors=auto_link_priors)
        self.lens_galaxies = lens_galaxies or []
        self.source_galaxies = source_galaxies or []

    class Analysis(PhaseImaging.Analysis):

        def __init__(self, lens_data, cosmology, phase_name, previous_results=None):

            super(LensSourcePlanePhase.Analysis, self).__init__(lens_data=lens_data, cosmology=cosmology,
                                                                phase_name=phase_name,
                                                                previous_results=previous_results)

        def tracer_for_instance(self, instance):
            return ray_tracing.TracerImageSourcePlanes(lens_galaxies=instance.lens_galaxies,
                                                       source_galaxies=instance.source_galaxies,
                                                       image_plane_grid_stack=self.lens_data.grid_stack,
                                                       border=self.lens_data.border, cosmology=self.cosmology)

        def padded_tracer_for_instance(self, instance):
            return ray_tracing.TracerImageSourcePlanes(lens_galaxies=instance.lens_galaxies,
                                                       source_galaxies=instance.source_galaxies,
                                                       image_plane_grid_stack=self.lens_data.padded_grid_stack,
                                                       cosmology=self.cosmology)

        @classmethod
        def log(cls, instance):
            logger.debug(
                "\nRunning lens/source lens for... \n\nLens Galaxy:\n{}\n\nSource Galaxy:\n{}\n\n".format(
                    instance.lens_galaxies, instance.source_galaxies))

    class Result(PhaseImaging.Result):

        def __init__(self, constant, figure_of_merit, variable, analysis, optimizer):
            """
            The result of a phase
            """

            super(LensSourcePlanePhase.Result, self).__init__(constant=constant, figure_of_merit=figure_of_merit,
                                                              variable=variable, analysis=analysis, optimizer=optimizer)

        @property
        def unmasked_lens_plane_model_image(self):
            return self.most_likely_fit.unmasked_model_image_of_planes[0]

        @property
        def unmasked_source_plane_model_image(self):
            return self.most_likely_fit.unmasked_model_image_of_planes[1]


class MultiPlanePhase(PhaseImaging):
    """
    Fit a simple source and lens system.
    """

    galaxies = PhasePropertyCollection("galaxies")

    @property
    def phase_property_collections(self):
        return [self.galaxies]

    def __init__(self, phase_name, galaxies=None, optimizer_class=non_linear.MultiNest,
                 sub_grid_size=2, image_psf_shape=None, use_positions=False, mask_function=None,
                 inner_circular_mask_radii=None, cosmology=cosmo.Planck15, auto_link_priors=False):
        """
        A phase with a simple source/lens model

        Parameters
        ----------
        lens_galaxies : [g.Galaxy] | [gm.GalaxyModel]
            A galaxy that acts as a gravitational lens
        source_galaxies: [g.Galaxy] | [gm.GalaxyModel]
            A galaxy that is being lensed
        optimizer_class: class
            The class of a non-linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        super(MultiPlanePhase, self).__init__(optimizer_class=optimizer_class,
                                              sub_grid_size=sub_grid_size,
                                              image_psf_shape=image_psf_shape,
                                              use_positions=use_positions,
                                              mask_function=mask_function,
                                              inner_circular_mask_radii=inner_circular_mask_radii,
                                              cosmology=cosmology,
                                              phase_name=phase_name,
                                              auto_link_priors=auto_link_priors)
        self.galaxies = galaxies

    class Analysis(PhaseImaging.Analysis):

        def __init__(self, lens_data, cosmology, phase_name, previous_results=None):
            self.lens_data = lens_data
            super(MultiPlanePhase.Analysis, self).__init__(lens_data=lens_data, cosmology=cosmology,
                                                           phase_name=phase_name, previous_results=previous_results)

        def tracer_for_instance(self, instance):
            return ray_tracing.TracerMultiPlanes(galaxies=instance.galaxies,
                                                 image_plane_grid_stack=self.lens_data.grid_stack,
                                                 border=self.lens_data.border, cosmology=self.cosmology)

        def padded_tracer_for_instance(self, instance):
            return ray_tracing.TracerMultiPlanes(galaxies=instance.galaxies,
                                                 image_plane_grid_stack=self.lens_data.padded_grid_stack,
                                                 cosmology=self.cosmology)

        @classmethod
        def log(cls, instance):
            logger.debug("\nRunning multi-plane for... \n\nGalaxies:\n{}\n\n".format(instance.galaxies))


class GalaxyFitPhase(AbstractPhase):

    galaxies = PhasePropertyCollection("galaxies")

    def __init__(self, phase_name, galaxies=None, use_intensities=False, use_surface_density=False, use_potential=False,
                 use_deflections=False, optimizer_class=non_linear.MultiNest, sub_grid_size=2,
                 mask_function=None, cosmology=cosmo.Planck15):
        """
        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper
        passed to it.

        Parameters
        ----------
        galaxy_data_class: class<gd.GalaxyData>
        optimizer_class: class
            The class of a non_linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        super(GalaxyFitPhase, self).__init__(optimizer_class=optimizer_class, cosmology=cosmology,
                                             phase_name=phase_name)
        self.use_intensities = use_intensities
        self.use_surface_density = use_surface_density
        self.use_potential = use_potential
        self.use_deflections = use_deflections
        self.galaxies = galaxies
        self.sub_grid_size = sub_grid_size
        self.mask_function = mask_function

    def run(self, galaxy_data, previous_results=None, mask=None):
        """
        Run this phase.

        Parameters
        ----------
        mask: Mask
            The default masks passed in by the pipeline
        noise_map
        array
        previous_results: ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper.
        """
        analysis = self.make_analysis(galaxy_data=galaxy_data, previous_results=previous_results, mask=mask)
        result = self.run_analysis(analysis)

        return self.make_result(result, analysis)

    def make_analysis(self, galaxy_data, previous_results=None, mask=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        mask: Mask
            The default masks passed in by the pipeline
        array
        noise_map
        previous_results: ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens: Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """

        mask = setup_phase_mask(data=galaxy_data[0], mask=mask, mask_function=self.mask_function,
                                inner_circular_mask_radii=None)

        self.pass_priors(previous_results)

        if self.use_intensities or self.use_surface_density or self.use_potential:

            galaxy_data = gd.GalaxyFitData(galaxy_data=galaxy_data[0], mask=mask, sub_grid_size=self.sub_grid_size,
                                           use_intensities=self.use_intensities,
                                           use_surface_density=self.use_surface_density,
                                           use_potential=self.use_potential,
                                           use_deflections_y=self.use_deflections,
                                           use_deflections_x=self.use_deflections)

            return self.__class__.AnalysisSingle(galaxy_data=galaxy_data, phase_name=self.phase_name,
                                                 cosmology=self.cosmology, previous_results=previous_results)

        elif self.use_deflections:

            galaxy_data_y = gd.GalaxyFitData(galaxy_data=galaxy_data[0], mask=mask, sub_grid_size=self.sub_grid_size,
                                             use_intensities=self.use_intensities,
                                             use_surface_density=self.use_surface_density,
                                             use_potential=self.use_potential,
                                             use_deflections_y=self.use_deflections, use_deflections_x=False)

            galaxy_data_x = gd.GalaxyFitData(galaxy_data=galaxy_data[1], mask=mask, sub_grid_size=self.sub_grid_size,
                                             use_intensities=self.use_intensities,
                                             use_surface_density=self.use_surface_density,
                                             use_potential=self.use_potential,
                                             use_deflections_y=False, use_deflections_x=self.use_deflections)

            return self.__class__.AnalysisDeflections(galaxy_data_y=galaxy_data_y, galaxy_data_x=galaxy_data_x,
                                                      cosmology=self.cosmology, phase_name=self.phase_name,
                                                      previous_results=previous_results)

    class Analysis(Phase.Analysis):

        def __init__(self, cosmology, phase_name, previous_results):

            super(GalaxyFitPhase.Analysis, self).__init__(cosmology=cosmology, phase_name=phase_name,
                                                          previous_results=previous_results)

            self.plot_galaxy_fit_all_at_end_png = \
                conf.instance.general.get('output', 'plot_galaxy_fit_all_at_end_png', bool)
            self.plot_galaxy_fit_all_at_end_fits = \
                conf.instance.general.get('output', 'plot_galaxy_fit_all_at_end_fits', bool)
            self.plot_galaxy_fit_as_subplot = \
                conf.instance.general.get('output', 'plot_galaxy_fit_as_subplot', bool)
            self.plot_galaxy_fit_image = \
                conf.instance.general.get('output', 'plot_galaxy_fit_image', bool)
            self.plot_galaxy_fit_noise_map = \
                conf.instance.general.get('output', 'plot_galaxy_fit_noise_map', bool)
            self.plot_galaxy_fit_model_image = \
                conf.instance.general.get('output', 'plot_galaxy_fit_model_image', bool)
            self.plot_galaxy_fit_residual_map = \
                conf.instance.general.get('output', 'plot_galaxy_fit_residual_map', bool)
            self.plot_galaxy_fit_chi_squared_map = \
                conf.instance.general.get('output', 'plot_galaxy_fit_chi_squared_map', bool)

        @classmethod
        def log(cls, instance):
            logger.debug(
                "\nRunning galaxy fit for... \n\nGalaxies::\n{}\n\n".format(instance.galaxies))

    # noinspection PyAbstractClass
    class AnalysisSingle(Analysis):

        def __init__(self, galaxy_data, cosmology, phase_name, previous_results=None):
            super(GalaxyFitPhase.AnalysisSingle, self).__init__(cosmology=cosmology, phase_name=phase_name,
                                                                previous_results=previous_results)

            self.galaxy_data = galaxy_data

        def fit(self, instance):
            fit = self.fit_for_instance(instance=instance)
            return fit.figure_of_merit

        def visualize(self, instance, suffix, during_analysis):

            self.plot_count += 1
            fit = self.fit_for_instance(instance=instance)

            if self.plot_galaxy_fit_as_subplot:

                galaxy_fit_plotters.plot_fit_subplot(
                    fit=fit, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                    units=self.plot_units,
                    output_path=self.output_image_path, output_format='png')

            if during_analysis:

                galaxy_fit_plotters.plot_fit_individuals(
                    fit=fit, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                    should_plot_image=self.plot_galaxy_fit_image,
                    should_plot_noise_map=self.plot_galaxy_fit_noise_map,
                    should_plot_model_image=self.plot_galaxy_fit_model_image,
                    should_plot_residual_map=self.plot_galaxy_fit_residual_map,
                    should_plot_chi_squared_map=self.plot_galaxy_fit_chi_squared_map,
                    units=self.plot_units,
                    output_path=self.output_image_path, output_format='png')

            elif not during_analysis:

                if self.plot_ray_tracing_all_at_end_png:

                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=self.output_image_path, output_format='png')

                if self.plot_ray_tracing_all_at_end_fits:

                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=self.output_fits_path, output_format='fits')

            return fit

        def fit_for_instance(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit: Fit
                A fractional value indicating how well this model fit and the model lens_data itself
            """
            return galaxy_fit.GalaxyFit(galaxy_data=self.galaxy_data, model_galaxies=instance.galaxies)

    # noinspection PyAbstractClass
    class AnalysisDeflections(Analysis):

        def __init__(self, galaxy_data_y, galaxy_data_x, cosmology, phase_name, previous_results=None):
            super(GalaxyFitPhase.AnalysisDeflections, self).__init__(cosmology=cosmology, phase_name=phase_name,
                                                                     previous_results=previous_results)

            self.galaxy_data_y = galaxy_data_y
            self.galaxy_data_x = galaxy_data_x

            self.output_image_y_path = "{}/image/fit_y_".format(self.phase_output_path)
            self.output_fits_y_path = "{}/image/fits/fit_y".format(self.phase_output_path)
            self.output_image_x_path = "{}/image/fit_x_".format(self.phase_output_path)
            self.output_fits_x_path = "{}/image/fits/fit_x".format(self.phase_output_path)

        def fit(self, instance):
            fit_y, fit_x = self.fit_for_instance(instance=instance)
            return fit_y.figure_of_merit + fit_x.figure_of_merit

        def visualize(self, instance, suffix, during_analysis):


            self.plot_count += 1
            fit_y, fit_x = self.fit_for_instance(instance=instance)

            if self.plot_galaxy_fit_as_subplot:

                galaxy_fit_plotters.plot_fit_subplot(
                    fit=fit_y, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                    units=self.plot_units,
                    output_path=self.output_image_y_path, output_format='png')

                galaxy_fit_plotters.plot_fit_subplot(
                    fit=fit_x, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                    units=self.plot_units,
                    output_path=self.output_image_x_path, output_format='png')

            if during_analysis:

                galaxy_fit_plotters.plot_fit_individuals(
                    fit=fit_y, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                    should_plot_image=self.plot_galaxy_fit_image,
                    should_plot_noise_map=self.plot_galaxy_fit_noise_map,
                    should_plot_model_image=self.plot_galaxy_fit_model_image,
                    should_plot_residual_map=self.plot_galaxy_fit_residual_map,
                    should_plot_chi_squared_map=self.plot_galaxy_fit_chi_squared_map,
                    units=self.plot_units,
                    output_path=self.output_image_y_path, output_format='png')

                galaxy_fit_plotters.plot_fit_individuals(
                    fit=fit_x, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                    should_plot_image=self.plot_galaxy_fit_image,
                    should_plot_noise_map=self.plot_galaxy_fit_noise_map,
                    should_plot_model_image=self.plot_galaxy_fit_model_image,
                    should_plot_residual_map=self.plot_galaxy_fit_residual_map,
                    should_plot_chi_squared_map=self.plot_galaxy_fit_chi_squared_map,
                    units=self.plot_units,
                    output_path=self.output_image_x_path, output_format='png')

            elif not during_analysis:

                if self.plot_ray_tracing_all_at_end_png:

                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit_y, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=self.output_image_y_path, output_format='png')

                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit_x, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=self.output_image_x_path, output_format='png')

                if self.plot_ray_tracing_all_at_end_fits:

                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit_y, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=self.output_fits_y_path, output_format='fits')

                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit_x, should_plot_mask=self.should_plot_mask, zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=self.output_fits_x_path, output_format='fits')

            return fit_y, fit_x

        def fit_for_instance(self, instance):

            fit_y = galaxy_fit.GalaxyFit(galaxy_data=self.galaxy_data_y, model_galaxies=instance.galaxies)
            fit_x = galaxy_fit.GalaxyFit(galaxy_data=self.galaxy_data_x, model_galaxies=instance.galaxies)

            return fit_y, fit_x

    class Result(Phase.Result):

        def __init__(self, constant, figure_of_merit, variable, analysis, optimizer):
            """
            The result of a phase
            """

            super(GalaxyFitPhase.Result, self).__init__(constant=constant, figure_of_merit=figure_of_merit,
                                                        variable=variable, analysis=analysis, optimizer=optimizer)


class SensitivityPhase(PhaseImaging):

    lens_galaxies = PhasePropertyCollection("lens_galaxies")
    source_galaxies = PhasePropertyCollection("source_galaxies")
    sensitive_galaxies = PhasePropertyCollection("sensitive_galaxies")

    def __init__(self, phase_name, lens_galaxies=None, source_galaxies=None, sensitive_galaxies=None,
                 optimizer_class=non_linear.MultiNest, sub_grid_size=2, mask_function=None,
                 cosmology=cosmo.Planck15):
        """
        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        super(SensitivityPhase, self).__init__(optimizer_class=optimizer_class, sub_grid_size=sub_grid_size,
                                               mask_function=mask_function, cosmology=cosmology,
                                               phase_name=phase_name)

        self.lens_galaxies = lens_galaxies or []
        self.source_galaxies = source_galaxies or []
        self.sensitive_galaxies = sensitive_galaxies or []

    # noinspection PyAbstractClass
    class Analysis(PhaseImaging.Analysis):

        def __init__(self, lens_data, cosmology, phase_name, previous_results=None):
            self.lens_data = lens_data
            super(PhaseImaging.Analysis, self).__init__(cosmology=cosmology, phase_name=phase_name,
                                                        previous_results=previous_results)

        def fit(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit: Fit
                A fractional value indicating how well this model fit and the model lens_data itself
            """
            tracer_normal = self.tracer_normal_for_instance(instance)
            tracer_sensitive = self.tracer_sensitive_for_instance(instance)
            fit = self.fit_for_tracers(tracer_normal=tracer_normal, tracer_sensitive=tracer_sensitive)
            return fit.figure_of_merit

        def visualize(self, instance, suffix, during_analysis):

            self.plot_count += 1

            tracer_normal = self.tracer_normal_for_instance(instance)
            tracer_sensitive = self.tracer_sensitive_for_instance(instance)
            fit = self.fit_for_tracers(tracer_normal=tracer_normal, tracer_sensitive=tracer_sensitive)

            ccd_plotters.plot_ccd_subplot(ccd_data=self.lens_data.ccd_data, mask=self.lens_data.mask,
                                          positions=self.lens_data.positions,
                                          output_path=self.output_image_path, output_format='png')

            ccd_plotters.plot_ccd_individual(ccd_data=self.lens_data.ccd_data, mask=self.lens_data.mask,
                                             positions=self.lens_data.positions,
                                             output_path=self.output_image_path,
                                             output_format='png')

            ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer_normal, output_path=self.output_image_path,
                                                          output_format='png', output_filename='tracer_normal')

            ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer_sensitive, output_path=self.output_image_path,
                                                          output_format='png', output_filename='tracer_sensitive')

            sensitivity_fit_plotters.plot_fit_subplot(fit=fit, output_path=self.output_image_path, output_format='png')

            return fit

        def tracer_normal_for_instance(self, instance):
            return ray_tracing.TracerImageSourcePlanes(lens_galaxies=instance.lens_galaxies,
                                                       source_galaxies=instance.source_galaxies,
                                                       image_plane_grid_stack=self.lens_data.grid_stack,
                                                       border=self.lens_data.border)

        def tracer_sensitive_for_instance(self, instance):
            return ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=instance.lens_galaxies + instance.sensitive_galaxies,
                source_galaxies=instance.source_galaxies,
                image_plane_grid_stack=self.lens_data.grid_stack,
                border=self.lens_data.border)

        def fit_for_tracers(self, tracer_normal, tracer_sensitive):
            return sensitivity_fit.fit_lens_data_with_sensitivity_tracers(lens_data=self.lens_data,
                                                                          tracer_normal=tracer_normal,
                                                                          tracer_sensitive=tracer_sensitive)

        @classmethod
        def log(cls, instance):
            logger.debug(
                "\nRunning lens/source lens for... \n\nLens Galaxy:\n{}\n\nSource Galaxy:\n{}\n\n Sensitive "
                "Galaxy\n{}\n\n "
                "".format(instance.lens_galaxies, instance.source_galaxies, instance.sensitive_galaxies))


class HyperAnalysis(object):

    def __init__(self):

        pass


def make_path_if_does_not_exist(path):
    if not os.path.exists(path):
        os.makedirs(path)



import numpy as np
import pytest

from autofit.tools import fit_util
from autolens.data import ccd
from autolens.data.array import scaled_array
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy as g
from autolens.lens.util import lens_fit_util as util
from autolens.lens import ray_tracing, lens_fit
from autolens.lens import lens_data as ld
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations
from autolens.model.inversion import regularization
from autolens.model.inversion import inversions
from test.mock.mock_profiles import MockLightProfile
from test.mock.mock_lens import MockTracer
from test.mock.mock_galaxy import MockHyperGalaxy

@pytest.fixture(name='lens_data_blur')
def make_lens_data_blur():

    image = np.array([[0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0]])
    psf = ccd.PSF(array=(np.array([[1.0, 1.0, 1.0],
                                     [1.0, 1.0, 1.0],
                                     [1.0, 1.0, 1.0]])), pixel_scale=1.0, renormalize=False)
    ccd_data = ccd.CCDData(image, pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

    mask = np.array([[True, True, True, True],
                   [True, False, False, True],
                   [True, False, False, True],
                   [True, True, True, True]])
    mask = msk.Mask(array=mask, pixel_scale=1.0)

    return ld.LensData(ccd_data, mask, sub_grid_size=1)

@pytest.fixture(name='lens_data_manual')
def make_li_manual():
    image = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 2.0, 3.0, 0.0],
                   [0.0, 4.0, 5.0, 6.0, 0.0],
                   [0.0, 7.0, 8.0, 9.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0, 0.0]])
    psf = ccd.PSF(array=(np.array([[1.0, 5.0, 9.0],
                                     [2.0, 5.0, 1.0],
                                     [3.0, 4.0, 0.0]])), pixel_scale=1.0)
    image = ccd.CCDData(image, pixel_scale=1.0, psf=psf, noise_map=np.ones((5, 5)))
    mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, True, True, True, True]]), pixel_scale=1.0)

    return ld.LensData(image, mask, sub_grid_size=1)

@pytest.fixture(name='hyper')
def make_hyper():
    class Hyper(object):

        def __init__(self):
            pass

    hyper = Hyper()

    hyper.hyper_model_image = np.array([[1.0, 3.0, 5.0, 7.0],
                                        [7.0, 9.0, 8.0, 1.0],
                                        [6.0, 4.0, 0.0, 9.0],
                                        [3.0, 4.0, 5.0, 6.0]])
    hyper.hyper_galaxy_images = [np.array([[1.0, 3.0, 5.0, 4.0],
                                           [7.0, 9.0, 8.0, 9.0],
                                           [6.0, 4.0, 0.0, 3.0],
                                           [6.0, 2.0, 3.0, 2.0]]),
                                 np.array([[1.0, 3.0, 5.0, 1.0],
                                           [7.0, 9.0, 8.0, 2.0],
                                           [6.0, 4.0, 0.0, 3.0],
                                           [1.0, 3.0, 4.0, 1.0]])]
    hyper.hyper_minimum_values = [0.2, 0.8]
    return hyper

@pytest.fixture(name='lens_data_hyper_no_blur')
def make_li_hyper_no_blur(hyper):

    image = np.array([[0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0]])

    psf = ccd.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                     [0.0, 1.0, 0.0],
                                     [0.0, 0.0, 0.0]])), pixel_scale=1.0, renormalize=False)
    ccd_data = ccd.CCDData(image, pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

    mask = np.array([[True, True, True, True],
                   [True, False, False, True],
                   [True, False, False, True],
                   [True, True, True, True]])
    mask = msk.Mask(array=mask, pixel_scale=1.0)

    return ld.LensDataHyper(ccd_data, mask, hyper_model_image=hyper.hyper_model_image,
                            hyper_galaxy_images=hyper.hyper_galaxy_images,
                            hyper_minimum_values=hyper.hyper_minimum_values, sub_grid_size=1)

@pytest.fixture(name='lens_data_hyper_manual')
def make_li_hyper_manual(hyper):
    image = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 2.0, 3.0, 0.0],
                   [0.0, 4.0, 5.0, 6.0, 0.0],
                   [0.0, 7.0, 8.0, 9.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0, 0.0]])
    psf = ccd.PSF(array=(np.array([[1.0, 5.0, 9.0],
                                     [2.0, 5.0, 1.0],
                                     [3.0, 4.0, 0.0]])), pixel_scale=1.0)
    ccd_data = ccd.CCDData(image, pixel_scale=1.0, psf=psf, noise_map=np.ones((5, 5)))
    mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, False, False, False, True],
                                   [True, True, True, True, True]]), pixel_scale=1.0)

    return ld.LensDataHyper(ccd_data, mask, hyper_model_image=hyper.hyper_model_image,
                            hyper_galaxy_images=hyper.hyper_galaxy_images,
                            hyper_minimum_values=hyper.hyper_minimum_values, sub_grid_size=1)


class TestAbstractLensFit:

    class TestAbstractLogic:

        def test__logic_in_abstract_fit(self, lens_data_manual):

            galaxy_light = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light],
                                                  image_plane_grid_stack=lens_data_manual.grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=None, psf=lens_data_manual.psf,
                                           map_to_scaled_array=lens_data_manual.map_to_scaled_array)

            assert fit.total_inversions == 0

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g.Galaxy(pixelization=pixelizations.Rectangular(),
                                                                          regularization=regularization.Constant())],
                                                  image_plane_grid_stack=lens_data_manual.grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=None, psf=lens_data_manual.psf,
                                           map_to_scaled_array=lens_data_manual.map_to_scaled_array)

            assert fit.total_inversions == 1

    class TestUnmaskedModelImage:

        def test__padded_tracer_is_none__unmasked_model_images_return_none(self, lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=None, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert fit.unmasked_model_image == None

        def test__padded_tracer_input__all_planes_have_light_profiles__unmasked_model_image_returns_array(self,
                                                                                                          lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)
            padded_tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image) == scaled_array.ScaledSquarePixelArray

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g0], source_galaxies=[g0, g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image) == scaled_array.ScaledSquarePixelArray

        def test__galaxy_in_tracer_has_pixelization__unmasked_model_image_is_none(self, lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0),
                          pixelization=pixelizations.Rectangular(), regularization=regularization.Constant())

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)
            padded_tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert fit.unmasked_model_image == None

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g0], source_galaxies=[g0, g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert fit.unmasked_model_image == None

    class TestUnmaskedModelImageOfPlanes:

        def test__padded_tracer_is_none__unmasked_model_images_return_none(self, lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=None, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert fit.unmasked_model_image_of_planes == None

        def test__padded_tracer_input__all_planes_have_light_profiles__unmasked_model_image_returns_array(self,
                                                                                                          lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)
            padded_tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image_of_planes[0]) == scaled_array.ScaledSquarePixelArray

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g0], source_galaxies=[g0, g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image_of_planes[0]) == scaled_array.ScaledSquarePixelArray
            assert type(fit.unmasked_model_image_of_planes[1]) == scaled_array.ScaledSquarePixelArray

        def test__galaxy_in_tracer_has_pixelization__unmasked_model_image_of_that_plane_is_none(self, lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0),
                          pixelization=pixelizations.Rectangular(), regularization=regularization.Constant())

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)

            padded_tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert fit.unmasked_model_image_of_planes[0] == None

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g0], source_galaxies=[g.Galaxy()],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert fit.unmasked_model_image_of_planes[0] == None
            assert type(fit.unmasked_model_image_of_planes[1]) == scaled_array.ScaledSquarePixelArray

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image_of_planes[0]) == scaled_array.ScaledSquarePixelArray
            assert fit.unmasked_model_image_of_planes[1] == None

    class TestUnmaskedModelImageOfPlanesAndGalaxies:

        def test__padded_tracer_is_none__unmasked_model_images_return_none(self, lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=None, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert fit.unmasked_model_image_of_planes_and_galaxies == None

        def test__padded_tracer_input__all_planes_have_light_profiles__unmasked_model_images_returns_array(self,
                                                                                                          lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)
            padded_tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0, g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image_of_planes_and_galaxies[0][0]) == scaled_array.ScaledSquarePixelArray
            assert type(fit.unmasked_model_image_of_planes_and_galaxies[0][1]) == scaled_array.ScaledSquarePixelArray

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g0], source_galaxies=[g0, g0, g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image_of_planes_and_galaxies[0][0]) == scaled_array.ScaledSquarePixelArray
            assert type(fit.unmasked_model_image_of_planes_and_galaxies[0][1]) == scaled_array.ScaledSquarePixelArray
            assert type(fit.unmasked_model_image_of_planes_and_galaxies[1][0]) == scaled_array.ScaledSquarePixelArray
            assert type(fit.unmasked_model_image_of_planes_and_galaxies[1][1]) == scaled_array.ScaledSquarePixelArray
            assert type(fit.unmasked_model_image_of_planes_and_galaxies[1][2]) == scaled_array.ScaledSquarePixelArray

        def test__galaxy_in_tracer_has_pixelization__unmasked_model_image_of_that_galaxy_is_none(self, lens_data_blur):

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            g_pix = g.Galaxy(pixelization=pixelizations.Rectangular(), regularization=regularization.Constant())

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data_blur.grid_stack)

            padded_tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g_pix],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert fit.unmasked_model_image_of_planes_and_galaxies[0][0] == None

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g_pix], source_galaxies=[g0],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image_of_planes_and_galaxies[0][0]) == scaled_array.ScaledSquarePixelArray
            assert fit.unmasked_model_image_of_planes_and_galaxies[0][1] == None
            assert type(fit.unmasked_model_image_of_planes_and_galaxies[1][0]) == scaled_array.ScaledSquarePixelArray

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy(), g_pix],
                                                                source_galaxies=[g0, g_pix],
                                                  image_plane_grid_stack=lens_data_blur.padded_grid_stack)

            fit = lens_fit.AbstractLensFit(tracer=tracer, padded_tracer=padded_tracer, psf=lens_data_blur.psf,
                                           map_to_scaled_array=lens_data_blur.map_to_scaled_array)

            assert type(fit.unmasked_model_image_of_planes_and_galaxies[0][0]) == scaled_array.ScaledSquarePixelArray
            assert fit.unmasked_model_image_of_planes_and_galaxies[0][1] == None
            assert type(fit.unmasked_model_image_of_planes_and_galaxies[1][0]) == scaled_array.ScaledSquarePixelArray
            assert fit.unmasked_model_image_of_planes_and_galaxies[1][1] == None


class TestAbstractLensProfileFit:

    class TestBlurredImage:

        def test__mock_tracer__2x2_image_all_1s__3x3_psf_all_1s__blurring_region__image_blurs_to_9s(self, lens_data_blur):

            tracer = MockTracer(unblurred_image_1d=lens_data_blur.mask.map_2d_array_to_masked_1d_array(lens_data_blur.image),
                                blurring_image_1d=np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]),
                                has_light_profile=True, has_hyper_galaxy=False, has_pixelization=False)

            fit = lens_fit.AbstractLensProfileFit(lens_data=lens_data_blur, tracer=tracer, padded_tracer=None)

            assert (fit.blurred_profile_image == np.array([[0.0, 0.0, 0.0, 0.0],
                                                           [0.0, 9.0, 9.0, 0.0],
                                                           [0.0, 9.0, 9.0, 0.0],
                                                           [0.0, 0.0, 0.0, 0.0]])).all()


class TestAbstractLensInversionFit:

    class TestModelImageOfPlanes:

        def test__model_images_are_none_and_an_image(self):

            image = np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 1.0, 1.0, 1.0, 0.0],
                           [0.0, 0.0, 0.0, 0.0, 0.0]])

            mask = np.array([[True, True, True, True, True],
                           [True, False, False, False, True],
                           [True, False, False, False, True],
                           [True, False, False, False, True],
                           [True, True, True, True, True]])

            mask = msk.Mask(mask, pixel_scale=1.0)

            psf = ccd.PSF(array=np.array([[0.0, 0.0, 0.0],
                                            [0.0, 1.0, 0.0],
                                            [0.0, 0.0, 0.0]]), pixel_scale=1.0)
            ccd_data = ccd.CCDData(image=image, pixel_scale=1.0, psf=psf, noise_map=np.ones((5, 5)))
            lens_data = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=2)

            galaxy_pix = g.Galaxy(pixelization=pixelizations.Rectangular(shape=(3, 3)),
                                  regularization=regularization.Constant(coefficients=(1.0,)))
            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[galaxy_pix],
                                                         image_plane_grid_stack=lens_data.grid_stack, border=None)

            fit = lens_fit.AbstractLensInversionFit(lens_data=lens_data, noise_map_1d=lens_data.noise_map_1d,
                                                    tracer=tracer)

            assert fit.model_image_of_planes[0] == None
            assert fit.model_image_of_planes[1] == pytest.approx(np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                                            [0.0, 1.0, 1.0, 1.0, 0.0],
                                                                            [0.0, 1.0, 1.0, 1.0, 0.0],
                                                                            [0.0, 1.0, 1.0, 1.0, 0.0],
                                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), 1e-2)


class TestAbstractLensProfileInversionFit:

    class TestModelImagesOfPlanes:

        def test___model_images_of_planes_are_profile_and_inversion_images(self, lens_data_manual):

            galaxy_light = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))
            galaxy_pix = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_light], source_galaxies=[galaxy_pix],
                                                         image_plane_grid_stack=lens_data_manual.grid_stack, border=None)

            fit = lens_fit.AbstractLensProfileInversionFit(lens_data=lens_data_manual,
                                                           noise_map_1d=lens_data_manual.noise_map_1d,
                                                           tracer=tracer, padded_tracer=None)

            blurred_profile_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
                convolver=lens_data_manual.convolver_image)

            blurred_profile_image = lens_data_manual.map_to_scaled_array(array_1d=blurred_profile_image_1d)

            assert (fit.model_image_of_planes[0] == blurred_profile_image).all()

            blurred_profile_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
                convolver=lens_data_manual.convolver_image)

            profile_subtracted_image_1d = lens_data_manual.image_1d - blurred_profile_image_1d

            mapper = pix.mapper_from_grid_stack_and_border(grid_stack=lens_data_manual.grid_stack, border=None)

            inversion = inversions.inversion_from_image_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d, noise_map_1d=lens_data_manual.noise_map_1d,
                convolver=lens_data_manual.convolver_mapping_matrix, mapper=mapper, regularization=reg)

            assert (fit.model_image_of_planes[1] == inversion.reconstructed_data).all()


class TestLensProfileFit:

    class TestLikelihood:

        def test__image__tracing_fits_data_perfectly__no_psf_blurring__lh_is_noise_normalization(self):

            psf = ccd.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                             [0.0, 1.0, 0.0],
                                             [0.0, 0.0, 0.0]])), pixel_scale=1.0)

            ccd_data = ccd.CCDData(image=np.ones((3, 3)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 3)))

            mask = msk.Mask(array=np.array([[True, True, True],
                                           [True, False, True],
                                           [True, True, True]]), pixel_scale=1.0)
            lens_data = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data.grid_stack)

            fit = lens_fit.LensProfileFit(lens_data=lens_data, tracer=tracer)
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

        def test__1x2_image__tracing_fits_data_with_chi_sq_5(self):

            psf = ccd.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                             [0.0, 1.0, 0.0],
                                             [0.0, 0.0, 0.0]])), pixel_scale=1.0)

            ccd_data = ccd.CCDData(5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 4)))
            ccd_data.image[1,2]  = 4.0

            mask = msk.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)

            lens_data = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0, size=2))
            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0], image_plane_grid_stack=lens_data.grid_stack)

            fit = lens_fit.LensProfileFit(lens_data=lens_data, tracer=tracer)

            assert fit.chi_squared == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.likelihood == -0.5 * (25.0 + 2.0*np.log(2 * np.pi * 1.0))

    class TestCompareToManual:

        def test___manual_image_and_psf(self, lens_data_manual):

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g0],
                                                         image_plane_grid_stack=lens_data_manual.grid_stack)

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0, g1], source_galaxies=[g0],
                                                                image_plane_grid_stack=lens_data_manual.padded_grid_stack)

            fit = lens_fit.fit_lens_data_with_tracer(lens_data=lens_data_manual, tracer=tracer,
                                                     padded_tracer=padded_tracer)

            assert lens_data_manual.noise_map == pytest.approx(fit.noise_map, 1e-4)

            model_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
                convolver=lens_data_manual.convolver_image)

            model_image = lens_data_manual.map_to_scaled_array(array_1d=model_image_1d)

            assert model_image == pytest.approx(fit.model_image, 1e-4)

            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=lens_data_manual.image,
                           mask=lens_data_manual.mask, model_data=model_image)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=lens_data_manual.mask, noise_map=lens_data_manual.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=lens_data_manual.mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(noise_map=lens_data_manual.noise_map,
                                                                                           mask=lens_data_manual.mask,)
            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)
            assert likelihood == fit.figure_of_merit

            blurred_image_of_planes = util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=tracer.total_planes, image_plane_image_1d_of_planes=tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_image_1d_of_planes,
            convolver=lens_data_manual.convolver_image, map_to_scaled_array=lens_data_manual.map_to_scaled_array)

            assert (blurred_image_of_planes[0] == fit.model_image_of_planes[0]).all()
            assert (blurred_image_of_planes[1] == fit.model_image_of_planes[1]).all()

            unmasked_blurred_image = \
                util.unmasked_blurred_image_from_padded_grid_stack_psf_and_unmasked_image(
                padded_grid_stack=lens_data_manual.padded_grid_stack, psf=lens_data_manual.psf,
                    unmasked_image_1d=padded_tracer.image_plane_image_1d)

            assert (unmasked_blurred_image == fit.unmasked_model_image).all()

            unmasked_blurred_image_of_planes = \
                util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(
                    planes=padded_tracer.planes, padded_grid_stack=lens_data_manual.padded_grid_stack, psf=lens_data_manual.psf)

            assert (unmasked_blurred_image_of_planes[0] == fit.unmasked_model_image_of_planes[0]).all()
            assert (unmasked_blurred_image_of_planes[1] == fit.unmasked_model_image_of_planes[1]).all()

            unmasked_blurred_image_of_galaxies = \
                util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(
                    planes=padded_tracer.planes, padded_grid_stack=lens_data_manual.padded_grid_stack, psf=lens_data_manual.psf)

            assert (unmasked_blurred_image_of_galaxies[0][0] == fit.unmasked_model_image_of_planes_and_galaxies[0][0]).all()
            assert (unmasked_blurred_image_of_galaxies[1][0] == fit.unmasked_model_image_of_planes_and_galaxies[1][0]).all()


class TestLensInversionFit:

    class TestCompareToManual:

        def test___manual_image_and_psf(self, lens_data_manual):

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))

            g0 = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g.Galaxy()], source_galaxies=[g0],
                                                         image_plane_grid_stack=lens_data_manual.grid_stack, border=None)

            fit = lens_fit.fit_lens_data_with_tracer(lens_data=lens_data_manual, tracer=tracer)

            mapper = pix.mapper_from_grid_stack_and_border(grid_stack=lens_data_manual.grid_stack, border=None)
            inversion = inversions.inversion_from_image_mapper_and_regularization(mapper=mapper,
                regularization=reg, image_1d=lens_data_manual.image_1d, noise_map_1d=lens_data_manual.noise_map_1d,
                convolver=lens_data_manual.convolver_mapping_matrix)

            assert inversion.reconstructed_data == pytest.approx(fit.model_image, 1e-4)

            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=lens_data_manual.image,
                           mask=lens_data_manual.mask, model_data=inversion.reconstructed_data)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                mask=lens_data_manual.mask, noise_map=lens_data_manual.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=lens_data_manual.mask)

            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=lens_data_manual.mask,
                                                                                  noise_map=lens_data_manual.noise_map)

            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = \
                util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                    chi_squared=chi_squared, regularization_term=inversion.regularization_term,
                    noise_normalization=noise_normalization)

            assert likelihood_with_regularization == pytest.approx(fit.likelihood_with_regularization, 1e-4)

            evidence = util.evidence_from_inversion_terms(chi_squared=chi_squared,
          regularization_term=inversion.regularization_term,
          log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
          log_regularization_term=inversion.log_det_regularization_matrix_term, noise_normalization=noise_normalization)

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit


class TestLensProfileInversionFit:

    class TestCompareToManual:

        def test___manual_image_and_psf(self, lens_data_manual):

            galaxy_light = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficients=(1.0,))
            galaxy_pix = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_light], source_galaxies=[galaxy_pix],
                                                         image_plane_grid_stack=lens_data_manual.grid_stack, border=None)

            fit = lens_fit.fit_lens_data_with_tracer(lens_data=lens_data_manual, tracer=tracer)

            blurred_profile_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
                convolver=lens_data_manual.convolver_image)

            blurred_profile_image = lens_data_manual.map_to_scaled_array(array_1d=blurred_profile_image_1d)

            assert blurred_profile_image == pytest.approx(fit.blurred_profile_image, 1e-4)

            profile_subtracted_image = lens_data_manual.image - blurred_profile_image

            assert profile_subtracted_image == pytest.approx(fit.profile_subtracted_image)

            blurred_profile_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
                convolver=lens_data_manual.convolver_image)

            profile_subtracted_image_1d = lens_data_manual.image_1d - blurred_profile_image_1d

            mapper = pix.mapper_from_grid_stack_and_border(grid_stack=lens_data_manual.grid_stack, border=None)

            inversion = inversions.inversion_from_image_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d, noise_map_1d=lens_data_manual.noise_map_1d,
                convolver=lens_data_manual.convolver_mapping_matrix, mapper=mapper, regularization=reg)

            model_image = blurred_profile_image + inversion.reconstructed_data

            assert model_image == pytest.approx(fit.model_image, 1e-4)

            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=lens_data_manual.image,
                           mask=lens_data_manual.mask, model_data=model_image)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                                                mask=lens_data_manual.mask, noise_map=lens_data_manual.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=lens_data_manual.mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=lens_data_manual.mask,
                                                                                  noise_map=lens_data_manual.noise_map)

            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = \
                util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                    chi_squared=chi_squared, regularization_term=inversion.regularization_term,
                    noise_normalization=noise_normalization)

            assert likelihood_with_regularization == pytest.approx(fit.likelihood_with_regularization, 1e-4)

            evidence = util.evidence_from_inversion_terms(chi_squared=chi_squared,
                                                                  regularization_term=inversion.regularization_term,
                                                                  log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                                                                  log_regularization_term=inversion.log_det_regularization_matrix_term, noise_normalization=noise_normalization)

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit


class TestLensProfileHyperFit:

    class TestLikelihood:

        def test__hyper_galaxy_adds_to_noise_normalization__chi_squared_is_0(self, lens_data_hyper_no_blur):
            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0, size=4))
            g1 = g.Galaxy(light_profile=MockLightProfile(value=0.0, size=4))

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0, g1],
                                                  image_plane_grid_stack=lens_data_hyper_no_blur.grid_stack)

            lens_data_hyper_no_blur.hyper_model_image_1d = np.array([1.0, 1.0, 1.0, 1.0])
            lens_data_hyper_no_blur.hyper_galaxy_images_1d = [np.array([1.0, 1.0, 1.0, 1.0]),
                                                           np.array([1.0, 1.0, 1.0, 1.0])]

            tracer.image_plane.galaxies[0].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=1.0,
                                                                          noise_power=1.0)
            tracer.image_plane.galaxies[1].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=2.0,
                                                                          noise_power=1.0)

            fit = lens_fit.LensProfileHyperFit(lens_data_hyper=lens_data_hyper_no_blur, tracer=tracer)

            chi_squared = 0.0
            noise_normalization = 4.0 * np.log(2 * np.pi * 4.0 ** 2.0)

            assert fit.likelihood == -0.5 * (chi_squared + noise_normalization)

        def test__hyper_galaxy_adds_to_noise_normalization_for_scaled_noise__chi_squared_nonzero(self,
                                                                                                 lens_data_hyper_no_blur):

            lens_data_hyper_no_blur.image[1:3,1:3] = 2.0

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0, size=4))
            g1 = g.Galaxy(light_profile=MockLightProfile(value=0.0, size=4))

            tracer = ray_tracing.TracerImagePlane(lens_galaxies=[g0, g1],
                                                  image_plane_grid_stack=lens_data_hyper_no_blur.grid_stack)

            lens_data_hyper_no_blur.hyper_model_image_1d = np.array([1.0, 1.0, 1.0, 1.0])
            lens_data_hyper_no_blur.hyper_galaxy_images_1d = [np.array([1.0, 1.0, 1.0, 1.0]),
                                                           np.array([1.0, 1.0, 1.0, 1.0])]


            tracer.image_plane.galaxies[0].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=1.0,
                                                                          noise_power=1.0)
            tracer.image_plane.galaxies[1].hyper_galaxy = MockHyperGalaxy(contribution_factor=0.0, noise_factor=2.0,
                                                                          noise_power=1.0)

            fit = lens_fit.LensProfileHyperFit(lens_data_hyper=lens_data_hyper_no_blur, tracer=tracer)

            chi_squared = 4.0 * (1.0 / (4.0)) ** 2.0
            noise_normalization = 4.0 * np.log(2 * np.pi * 4.0 ** 2.0)

            assert fit.likelihood == -0.5 * (chi_squared + noise_normalization)

    class TestCompareToManual:

        def test___manual_image_and_psf(self, lens_data_hyper_manual):

            hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0), hyper_galaxy=hyper_galaxy)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                         image_plane_grid_stack=lens_data_hyper_manual.grid_stack)

            padded_tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[g0], source_galaxies=[g0],
                                                               image_plane_grid_stack=lens_data_hyper_manual.padded_grid_stack)

            fit = lens_fit.hyper_fit_lens_data_with_tracer(lens_data_hyper=lens_data_hyper_manual, tracer=tracer,
                                                           padded_tracer=padded_tracer)

            contributions_1d = util.contribution_maps_1d_from_hyper_images_and_galaxies(
                hyper_model_image_1d=lens_data_hyper_manual.hyper_model_image_1d,
                hyper_galaxy_images_1d=lens_data_hyper_manual.hyper_galaxy_images_1d,
                hyper_galaxies=tracer.hyper_galaxies, hyper_minimum_values=lens_data_hyper_manual.hyper_minimum_values)

            contribution_maps = list(map(lambda contribution_1d :
                                     lens_data_hyper_manual.map_to_scaled_array(array_1d=contribution_1d),
                                     contributions_1d))

            assert contribution_maps[0] == pytest.approx(fit.contribution_maps[0], 1.0e-4)

            hyper_noise_map_1d = util.scaled_noise_map_from_hyper_galaxies_and_contribution_maps(
                contribution_maps=contributions_1d, hyper_galaxies=tracer.hyper_galaxies,
                noise_map=lens_data_hyper_manual.noise_map_1d)

            hyper_noise_map = lens_data_hyper_manual.map_to_scaled_array(array_1d=hyper_noise_map_1d)

            assert hyper_noise_map == pytest.approx(fit.noise_map, 1.0e-4)

            model_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
                convolver=lens_data_hyper_manual.convolver_image)

            model_image = lens_data_hyper_manual.map_to_scaled_array(array_1d=model_image_1d)

            assert model_image == pytest.approx(fit.model_image, 1e-4)

            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=lens_data_hyper_manual.image,
                                                                                   mask=lens_data_hyper_manual.mask,
                                                                                   model_data=model_image)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                                                mask=lens_data_hyper_manual.mask, noise_map=hyper_noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=lens_data_hyper_manual.mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=lens_data_hyper_manual.mask,
                                                                                  noise_map=lens_data_hyper_manual.noise_map)
            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.figure_of_merit, 1e-4)
            assert likelihood == pytest.approx(fit.figure_of_merit, 1e-4)

            blurred_image_of_planes = util.blurred_image_of_planes_from_1d_images_and_convolver(
            total_planes=tracer.total_planes, image_plane_image_1d_of_planes=tracer.image_plane_image_1d_of_planes,
            image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_image_1d_of_planes,
            convolver=lens_data_hyper_manual.convolver_image,
            map_to_scaled_array=lens_data_hyper_manual.map_to_scaled_array)

            assert (blurred_image_of_planes[0] == fit.model_image_of_planes[0]).all()
            assert (blurred_image_of_planes[1] == fit.model_image_of_planes[1]).all()

            unmasked_blurred_image = \
                util.unmasked_blurred_image_from_padded_grid_stack_psf_and_unmasked_image(
                padded_grid_stack=lens_data_hyper_manual.padded_grid_stack, psf=lens_data_hyper_manual.psf,
                    unmasked_image_1d=padded_tracer.image_plane_image_1d)

            assert (unmasked_blurred_image == fit.unmasked_model_image).all()

            unmasked_blurred_image_of_planes = \
                util.unmasked_blurred_image_of_planes_from_padded_grid_stack_and_psf(
                    planes=padded_tracer.planes, padded_grid_stack=lens_data_hyper_manual.padded_grid_stack,
                    psf=lens_data_hyper_manual.psf)

            assert (unmasked_blurred_image_of_planes[0] == fit.unmasked_model_image_of_planes[0]).all()
            assert (unmasked_blurred_image_of_planes[1] == fit.unmasked_model_image_of_planes[1]).all()

            unmasked_blurred_image_of_galaxies = \
                util.unmasked_blurred_image_of_planes_and_galaxies_from_padded_grid_stack_and_psf(
                    planes=padded_tracer.planes, padded_grid_stack=lens_data_hyper_manual.padded_grid_stack,
                    psf=lens_data_hyper_manual.psf)

            assert (unmasked_blurred_image_of_galaxies[0][0] == fit.unmasked_model_image_of_planes_and_galaxies[0][0]).all()
            assert (unmasked_blurred_image_of_galaxies[1][0] == fit.unmasked_model_image_of_planes_and_galaxies[1][0]).all()


class TestLensInversionHyperFit:

    class TestCompareToManual:

        def test___manual_image_and_psf(self, lens_data_hyper_manual):

            pix = pixelizations.Rectangular(shape=(3, 3))
            mapper = pix.mapper_from_grid_stack_and_border(grid_stack=lens_data_hyper_manual.grid_stack,
                                                           border=lens_data_hyper_manual.border)
            reg = regularization.Constant(coefficients=(1.0,))

            hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)
            hyp_galaxy = g.Galaxy(hyper_galaxy=hyper_galaxy)
            inv_galaxy = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[hyp_galaxy, hyp_galaxy],
                                                         source_galaxies=[inv_galaxy],
                                                         image_plane_grid_stack=lens_data_hyper_manual.grid_stack,
                                                         border=None)

            fit = lens_fit.hyper_fit_lens_data_with_tracer(lens_data_hyper=lens_data_hyper_manual, tracer=tracer)

            contributions_1d = util.contribution_maps_1d_from_hyper_images_and_galaxies(
                hyper_model_image_1d=lens_data_hyper_manual.hyper_model_image_1d,
                hyper_galaxy_images_1d=lens_data_hyper_manual.hyper_galaxy_images_1d,
                hyper_galaxies=tracer.hyper_galaxies, hyper_minimum_values=lens_data_hyper_manual.hyper_minimum_values)

            contribution_maps = list(map(lambda contribution_1d:
                                     lens_data_hyper_manual.map_to_scaled_array(array_1d=contribution_1d),
                                     contributions_1d))

            assert contribution_maps[0] == pytest.approx(fit.contribution_maps[0], 1.0e-4)

            hyper_noise_map_1d = util.scaled_noise_map_from_hyper_galaxies_and_contribution_maps(
                contribution_maps=contributions_1d, hyper_galaxies=tracer.hyper_galaxies,
                noise_map=lens_data_hyper_manual.noise_map_1d)

            hyper_noise_map = lens_data_hyper_manual.map_to_scaled_array(array_1d=hyper_noise_map_1d)

            assert hyper_noise_map == pytest.approx(fit.noise_map, 1.0e-4)

            inversion = inversions.inversion_from_image_mapper_and_regularization(mapper=mapper,
                                                                                  regularization=reg, image_1d=lens_data_hyper_manual.image_1d, noise_map_1d=hyper_noise_map_1d,
                                                                                  convolver=lens_data_hyper_manual.convolver_mapping_matrix)

            assert inversion.reconstructed_data == pytest.approx(fit.model_image, 1e-4)

            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=lens_data_hyper_manual.image,
                           mask=lens_data_hyper_manual.mask, model_data=inversion.reconstructed_data)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                                                mask=lens_data_hyper_manual.mask, noise_map=hyper_noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=lens_data_hyper_manual.mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=lens_data_hyper_manual.mask,
                                                                                  noise_map=hyper_noise_map)

            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = \
                util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                    chi_squared=chi_squared, regularization_term=inversion.regularization_term,
                    noise_normalization=noise_normalization)

            assert likelihood_with_regularization == pytest.approx(fit.likelihood_with_regularization, 1e-4)

            evidence = util.evidence_from_inversion_terms(chi_squared=chi_squared,
                                                                  regularization_term=inversion.regularization_term,
                                                                  log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                                                                  log_regularization_term=inversion.log_det_regularization_matrix_term, noise_normalization=noise_normalization)

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit


class TestLensProfileInversionHyperFit:

    class TestCompareToManual:

        def test___manual_image_and_psf(self, lens_data_hyper_manual):

            pix = pixelizations.Rectangular(shape=(3, 3))
            mapper = pix.mapper_from_grid_stack_and_border(grid_stack=lens_data_hyper_manual.grid_stack,
                                                           border=lens_data_hyper_manual.border)
            reg = regularization.Constant(coefficients=(1.0,))

            hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)
            hyp_galaxy = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0), hyper_galaxy=hyper_galaxy)
            inv_galaxy = g.Galaxy(pixelization=pix, regularization=reg)

            tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[hyp_galaxy, hyp_galaxy],
                                                         source_galaxies=[inv_galaxy],
                                                         image_plane_grid_stack=lens_data_hyper_manual.grid_stack,
                                                         border=None)

            fit = lens_fit.hyper_fit_lens_data_with_tracer(lens_data_hyper=lens_data_hyper_manual, tracer=tracer)

            contributions_1d = util.contribution_maps_1d_from_hyper_images_and_galaxies(
                hyper_model_image_1d=lens_data_hyper_manual.hyper_model_image_1d,
                hyper_galaxy_images_1d=lens_data_hyper_manual.hyper_galaxy_images_1d,
                hyper_galaxies=tracer.hyper_galaxies, hyper_minimum_values=lens_data_hyper_manual.hyper_minimum_values)

            contribution_maps = list(map(lambda contribution_1d:
                                     lens_data_hyper_manual.map_to_scaled_array(array_1d=contribution_1d),
                                     contributions_1d))

            assert contribution_maps[0] == pytest.approx(fit.contribution_maps[0], 1.0e-4)

            hyper_noise_map_1d = util.scaled_noise_map_from_hyper_galaxies_and_contribution_maps(
                contribution_maps=contributions_1d, hyper_galaxies=tracer.hyper_galaxies,
                noise_map=lens_data_hyper_manual.noise_map_1d)

            hyper_noise_map = lens_data_hyper_manual.map_to_scaled_array(array_1d=hyper_noise_map_1d)

            assert hyper_noise_map == pytest.approx(fit.noise_map, 1.0e-4)

            blurred_profile_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
                convolver=lens_data_hyper_manual.convolver_image)

            blurred_profile_image = lens_data_hyper_manual.map_to_scaled_array(array_1d=blurred_profile_image_1d)

            assert blurred_profile_image == pytest.approx(fit.blurred_profile_image, 1e-4)

            profile_subtracted_image = lens_data_hyper_manual.image - blurred_profile_image

            assert profile_subtracted_image == pytest.approx(fit.profile_subtracted_image)

            blurred_profile_image_1d = util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_image_1d, blurring_image_1d=tracer.image_plane_blurring_image_1d,
                convolver=lens_data_hyper_manual.convolver_image)

            profile_subtracted_image_1d = lens_data_hyper_manual.image_1d - blurred_profile_image_1d

            inversion = inversions.inversion_from_image_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d, noise_map_1d=hyper_noise_map_1d,
                convolver=lens_data_hyper_manual.convolver_mapping_matrix, mapper=mapper, regularization=reg)

            model_image = blurred_profile_image + inversion.reconstructed_data

            assert model_image == pytest.approx(fit.model_image, 1e-4)

            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=lens_data_hyper_manual.image,
                                                                                   mask=lens_data_hyper_manual.mask,
                                                                                   model_data=model_image)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                                                mask=lens_data_hyper_manual.mask, noise_map=hyper_noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=lens_data_hyper_manual.mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(noise_map=hyper_noise_map,
                                                                                           mask=lens_data_hyper_manual.mask)

            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                         noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = \
                util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                    chi_squared=chi_squared, regularization_term=inversion.regularization_term,
                    noise_normalization=noise_normalization)

            assert likelihood_with_regularization == pytest.approx(fit.likelihood_with_regularization, 1e-4)

            evidence = util.evidence_from_inversion_terms(chi_squared=chi_squared,
                                      regularization_term=inversion.regularization_term,
                                      log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                                      log_regularization_term=inversion.log_det_regularization_matrix_term,
                                      noise_normalization=noise_normalization)

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit


class MockTracerPositions:

    def __init__(self, positions, noise=None):
        self.positions = positions
        self.noise = noise


class TestPositionFit:

    def test__x1_positions__mock_position_tracer__maximum_separation_is_correct(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(2)

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(1.0) + np.square(3.0))

        tracer = MockTracerPositions(positions=[np.array([[-2.0, -4.0], [1.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(positions=[np.array([[8.0, 4.0], [-9.0, -4.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_positions__mock_position_tracer__maximum_separation_is_correct(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(positions=[np.array([[-2.0, -4.0], [1.0, 3.0], [0.1, 0.1], [-0.1, -0.1],
                                                          [0.3, 0.4], [-0.6, 0.5]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(positions=[np.array([[8.0, 4.0], [8.0, 4.0], [-9.0, -4.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_sets_of_positions__multiple_sets_of_max_distances(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                                                np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                                                np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)

        assert fit.maximum_separations[0] == 1.0
        assert fit.maximum_separations[1] == np.sqrt(18)
        assert fit.maximum_separations[2] == np.sqrt(18)

    def test__likelihood__is_sum_of_separations_divided_by_noise(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                                                np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                                                np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])])

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.chi_squared_map[0] == 1.0
        assert fit.chi_squared_map[1] == pytest.approx(18.0, 1e-4)
        assert fit.chi_squared_map[2] == pytest.approx(18.0, 1e-4)
        assert fit.figure_of_merit == pytest.approx(-0.5 * (1.0 + 18 + 18), 1e-4)

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=2.0)
        assert fit.chi_squared_map[0] == (1.0 / 2.0) ** 2.0
        assert fit.chi_squared_map[1] == pytest.approx((np.sqrt(18.0) / 2.0) ** 2.0, 1e-4)
        assert fit.chi_squared_map[2] == pytest.approx((np.sqrt(18.0) / 2.0) ** 2.0, 1e-4)
        assert fit.figure_of_merit == pytest.approx(-0.5 * ((1.0 / 2.0) ** 2.0 + (np.sqrt(18.0) / 2.0) ** 2.0 +
                                                            (np.sqrt(18.0) / 2.0) ** 2.0), 1e-4)

    def test__threshold__if_not_met_returns_ray_tracing_exception(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)

        assert fit.maximum_separation_within_threshold(threshold=100.0) == True
        assert fit.maximum_separation_within_threshold(threshold=0.1) == False



import numpy as np
import pytest

from autofit.tools import fit_util
from autolens.data.array import mask as msk, scaled_array as sca
from autolens.model.galaxy import galaxy_data as gd
from autolens.model.galaxy import galaxy as g, galaxy_fit
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.mock.mock_galaxy import MockGalaxy


class TestGalaxyFit:

    class TestLikelihood:

        def test__1x1_image__light_profile_fits_data_perfectly__lh_is_noise(self):

            image = sca.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)

            noise_map = sca.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)

            galaxy_data = gd.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

            mask = msk.Mask(array=np.array([[True, True, True],
                                           [True, False, True],
                                           [True, True, True]]), pixel_scale=1.0)
            g0 = MockGalaxy(value=1.0)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_intensities=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_surface_density=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_potential=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_deflections_y=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_deflections_x=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

        def test__1x2_image__noise_not_1__alls_correct(self):

            image = sca.ScaledSquarePixelArray(array=5.0*np.ones((3, 4)), pixel_scale=1.0)
            image[1,2] = 4.0

            noise_map = sca.ScaledSquarePixelArray(array=2.0*np.ones((3, 4)), pixel_scale=1.0)

            galaxy_data = gd.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

            mask = msk.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)

            g0 = MockGalaxy(value=1.0, shape=2)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_intensities=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])

            assert fit.model_galaxies == [g0]
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_surface_density=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_potential=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_deflections_y=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=1,
                                               use_deflections_x=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

    class TestCompareToManual:

        def test__intensities(self):

            image = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)

            noise_map = 2.0 * np.ones((5, 5))

            galaxy_data = gd.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=2, use_intensities=True)

            galaxy = g.Galaxy(light=lp.SphericalSersic(centre=(1.0, 2.0), intensity=1.0))
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

            assert fit.model_galaxies == [galaxy]

            model_data_1d = galaxy.intensities_from_grid(grid=galaxy_fit_data.grid_stack.sub)
            model_data_1d = galaxy_fit_data.grid_stack.sub.sub_data_to_regular_data(sub_array=model_data_1d)
            model_data = galaxy_fit_data.map_to_scaled_array(array_1d=model_data_1d)
            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=galaxy_fit_data.image, mask=galaxy_fit_data.mask,
                                                                               model_data=model_data)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map, 
                                                                                 mask=mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=galaxy_fit_data.mask,
                                                                                  noise_map=galaxy_fit_data.noise_map)
            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        def test__surface_density(self):

            image = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)
            noise_map = 2.0 * np.ones((5, 5))

            galaxy_data = gd.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=2, use_surface_density=True)

            galaxy = g.Galaxy(mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0))
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

            assert fit.model_galaxies == [galaxy]

            model_data_1d = galaxy.surface_density_from_grid(grid=galaxy_fit_data.grid_stack.sub)
            model_data_1d = galaxy_fit_data.grid_stack.sub.sub_data_to_regular_data(sub_array=model_data_1d)
            model_data = galaxy_fit_data.map_to_scaled_array(array_1d=model_data_1d)
            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=galaxy_fit_data.image, mask=galaxy_fit_data.mask,
                                                                               model_data=model_data)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=galaxy_fit_data.mask,
                                                                                  noise_map=galaxy_fit_data.noise_map)
            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)
            
        def test__potential(self):

            image = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)

            noise_map = 2.0 * np.ones((5, 5))

            galaxy_data = gd.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)

            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=2, use_potential=True)

            galaxy = g.Galaxy(mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0))

            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

            assert fit.model_galaxies == [galaxy]

            model_data_1d = galaxy.potential_from_grid(grid=galaxy_fit_data.grid_stack.sub)
            model_data_1d = galaxy_fit_data.grid_stack.sub.sub_data_to_regular_data(sub_array=model_data_1d)
            model_data = galaxy_fit_data.map_to_scaled_array(array_1d=model_data_1d)
            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=galaxy_fit_data.image, mask=galaxy_fit_data.mask,
                                                                               model_data=model_data)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=galaxy_fit_data.mask,
                                                                                  noise_map=galaxy_fit_data.noise_map)
            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        def test__deflections_y(self):

            image = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)

            noise_map = 2.0 * np.ones((5, 5))

            galaxy_data = gd.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)

            galaxy = g.Galaxy(mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0))
            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=2,
                                               use_deflections_y=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

            assert fit.model_galaxies == [galaxy]

            model_data_1d = galaxy.deflections_from_grid(grid=galaxy_fit_data.grid_stack.sub)
            model_data_1d = galaxy_fit_data.grid_stack.sub.sub_data_to_regular_data(sub_array=model_data_1d[:,0])
            model_data = galaxy_fit_data.map_to_scaled_array(array_1d=model_data_1d)
            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=galaxy_fit_data.image,
                                                                               mask=galaxy_fit_data.mask,
                                                                               model_data=model_data)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                        mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map, mask=mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=galaxy_fit_data.mask,
                                                                                           noise_map=galaxy_fit_data.noise_map)
            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        def test__deflections_x(self):

            image = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)

            noise_map = 2.0 * np.ones((5, 5))

            galaxy_data = gd.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)

            galaxy = g.Galaxy(mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0))
            galaxy_fit_data = gd.GalaxyFitData(galaxy_data=galaxy_data, mask=mask, sub_grid_size=2, use_deflections_x=True)
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

            assert fit.model_galaxies == [galaxy]

            model_data_1d = galaxy.deflections_from_grid(grid=galaxy_fit_data.grid_stack.sub)
            model_data_1d = galaxy_fit_data.grid_stack.sub.sub_data_to_regular_data(sub_array=model_data_1d[:,1])
            model_data = galaxy_fit_data.map_to_scaled_array(array_1d=model_data_1d)
            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=galaxy_fit_data.image, mask=galaxy_fit_data.mask,
                                                                               model_data=model_data)

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(residual_map=residual_map,
                                                                    mask=galaxy_fit_data.mask, noise_map=galaxy_fit_data.noise_map)

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                                 mask=mask)
            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(mask=galaxy_fit_data.mask,
                                                                                  noise_map=galaxy_fit_data.noise_map)
            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                          noise_normalization=noise_normalization)

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)
